#define PYGGI_MODULE
#include "pyggi.h"

pygii_C_api * gii_c_api;

static pyggi_C_api capi;

int
pyggi_getInt(PyObject *o, int *i) {
	int a;
	a=PyInt_AsLong(o);
	if(PyErr_Occurred())
		return -1;
	*i=a;
	return 0;
}

int
pyggi_getCoords(PyObject *o, ggi_coord *c) {
	int r[2];
	if(PySequence_Check(o)) {
		if(PySequence_Size(o)==2) {
			r[0]=PyInt_AsLong(PySequence_GetItem(o,0));
			r[1]=PyInt_AsLong(PySequence_GetItem(o,1));
			if(PyErr_Occurred())
				return -1;
			c->x=r[0];
			c->y=r[1];
			return 0;
		}
	}
	PyErr_SetString(PyExc_TypeError,"a 2-int sequence is required");
	return -1;
}

int
pyggi_getRect(PyObject *o, int i[4]) {
	int r[4];
	if(PySequence_Check(o)) {
		if(PySequence_Size(o)==4) {
			r[0]=PyInt_AsLong(PySequence_GetItem(o,0));
			r[1]=PyInt_AsLong(PySequence_GetItem(o,1));
			r[2]=PyInt_AsLong(PySequence_GetItem(o,2));
			r[3]=PyInt_AsLong(PySequence_GetItem(o,3));	     
			if(PyErr_Occurred())
				return -1;
			i[0]=r[0];
			i[1]=r[1];
			i[2]=r[2];
			i[3]=r[3];	
			return 0;
		}
	}
	
	PyErr_SetString(PyExc_TypeError,"a 4-int sequence is required");
	return -1;
}

int
pyggi_getPixel(PyObject *o, ggi_pixel *p, ggi_visual * v) {
	ggi_color c;
	float r,g,b;
	if(PyInt_Check(o)) {
		*p=PyInt_AsLong(o);
		return 0;
	}
	if(PySequence_Check(o)) {
		if(PySequence_Size(o)==3) {
			r=PyFloat_AsDouble(PySequence_GetItem(o,0));
			g=PyFloat_AsDouble(PySequence_GetItem(o,1));
			b=PyFloat_AsDouble(PySequence_GetItem(o,2));
			if(PyErr_Occurred())
				return -1;
			float2color(r,g,b,&c);
			*p=ggiMapColor(v,&c);
			return 0;
		}
	}
	PyErr_SetString(PyExc_TypeError,
			"an int or a 3-float sequence is required");
	return -1;
}

PyObject *
pyggi_newMode(ggi_mode m) {
	Mode *r;
	r = PyObject_New(Mode, &ModeType);
	if (r)
		r->mode = m;
	return (PyObject *)r;
}

PyObject *
pyggi_newPackFromLength(Visual * vis, int length) {
	Pack *r;
	
	r = PyObject_New(Pack, &PackType);
	if (r==NULL) {
		return NULL;
	}

	/* XXX see how to avoid duplicating Pack_init here */

	r -> vis    = vis;
	r -> length = length;
	r -> alloc  = ((r->length*GT_SIZE(vis->mode.graphtype))/8+1);
	r -> pixels = malloc(r->alloc);
	if(!r->pixels) {
		r-> vis = NULL;
		PyObject_Del(r);
		PyErr_SetString(PyExc_MemoryError, "out of memory");
		return NULL;
	}
	return (PyObject *)r;
}


PyObject *
pyggi_newPackFromSequence(Visual *vis, PyObject *a) {
	Pack *r;
	PyObject *o;
	ggi_color *c;
	int i;
	size_t l;
	double cr,cg,cb;

	l = (size_t)PySequence_Length(a);
	c = calloc(l, sizeof(ggi_color));
	if(!c) {
		PyErr_SetString(PyExc_MemoryError,"out of memory");
		return NULL;
	}
	for(i=0;i<l;i++) {
		o = PySequence_Fast_GET_ITEM(a,i);
		if(PySequence_Check(o))
			if(PySequence_Fast_GET_SIZE(o)==3) {
			 cr = PyFloat_AsDouble(PySequence_Fast_GET_ITEM(o,0));
			 cg = PyFloat_AsDouble(PySequence_Fast_GET_ITEM(o,1));
			 cb = PyFloat_AsDouble(PySequence_Fast_GET_ITEM(o,2));
			 float2color(cr,cg,cb,c+i);
			}
	}

	r = (Pack*)pyggi_newPackFromLength(vis, l);
	if(!r) {
		free(c);
		return NULL;
	}
	if(ggiPackColors(r->vis->vis, r->pixels, c, r->length)) {
		PyObject_Del(r);
		PyErr_SetString(PyExc_SystemError,
				"ggiPackColors failed");
		r = NULL;
	}
	free(c);
	return (PyObject *)r;
}


PyObject *
pyggi_newPackFromPack(Visual * vis, Pack * o) {
	Pack *r;
	ggi_color *c;

	r = NULL;	
	/* check if the visuals are the same */
	if (vis->vis == o->vis->vis) {
		/* duplicate */
		r = (Pack*)pyggi_newPackFromLength(vis, o->length);
		if(!r) return NULL;
		memmove(r->pixels, o->pixels, r->alloc);
	}
	else {
		/* unmap and remap */
		c = calloc(o->length, sizeof(ggi_color));
		if (!c) {
			PyErr_SetString(PyExc_MemoryError, "out of memory");
			return NULL;
		}
		if (ggiUnpackPixels(o->vis->vis,o->pixels,c,(int)o->length)) {
			free(c);
			PyErr_SetString(PyExc_SystemError,
					"ggiUnpackPixels failed");
			return NULL;
		}
		r = (Pack*)pyggi_newPackFromLength(vis, o->length);
		if(!r) {
			free(c);
			return NULL;
		}
		
		if(ggiPackColors(r->vis->vis, r->pixels, c, r->length)) {
			PyObject_Del(r);
			PyErr_SetString(PyExc_SystemError,
					"ggiPackColors failed");
			r = NULL;
		}
		
		free(c);
	}
	return (PyObject *)r;
}

/*********************************************************
 * Extension handling functions                          *
 *********************************************************/
int
pyggi_attachExtension(Visual         *vis,
		      char           *n,
		      ExtGetAttrFunc ga,
		      ExtSetAttrFunc sa) {
	Extension * e;
	
	/* test if it's not already attached */
	
	for(e=vis->exts;e;e=e->next)
		if(!strcmp(e->name,n))
			return -1;
	e=malloc(sizeof(Extension));
	if(e==NULL)
		return -2;
	
	e->next    = vis->exts;
	e->name    = n;
	e->getattr = ga;
	e->setattr = sa;
	vis->exts  = e;
	vis->nb_exts++;

	return 0;
}

int
pyggi_detachExtension(Visual *vis,
		      char   *name) {
	Extension ** t;
	Extension * e;
	
	for(e=vis->exts,t=&vis->exts;e;e=e->next,t=&e->next) {
		if(!strcmp(e->name,name)) {
			/* Have to remove this one*/
			*t=e->next;			
			free(e);
			vis->nb_exts--;
			return 0;
		}
	}
	/* Not registered */
	return -1;
}

void
pyggi_detachAll(Visual *vis) {
	Extension * r;
	while(vis->exts) {
		r=vis->exts->next;
		free(vis->exts);
		vis->exts=r;
	}
	vis->nb_exts=0;
}


PyObject *
pyggi_getExtensionList(Visual *vis) {
	PyObject  *r;
	Extension *e;
	int i;
	
	r=PyList_New(vis->nb_exts);
	if(r) {
		e=vis->exts;
		for(i=0;i<vis->nb_exts;i++,e=e->next)
			PyList_SET_ITEM(r,i,PyString_FromString(e->name));
	}
	return r;
}

static void importModules(void) {
	PyObject *gii;
	PyObject *giidict;
	PyObject *capiobj;
	
	gii= PyImport_ImportModule("GGI._giimodule");
	if (gii!= NULL) {
		giidict = PyModule_GetDict(gii);
		capiobj = PyDict_GetItemString(giidict, "_giimodule_C_api");
		if (PyCObject_Check(capiobj)) {
			gii_c_api=(pygii_C_api *)PyCObject_AsVoidPtr(capiobj);
		}
		/* 
		Py_DECREF(giidict);
		Py_DECREF(capiobj);
		*/
	} else {
		Py_FatalError("Can't import GGI._giimodule");
		return;
	}
}

#ifndef PyMODINIT_FUNC
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
init_ggimodule(void) {
	PyObject *m;

	/* Initialize the types */ 

	if(PyType_Ready(&ModeType)   < 0) return;
	if(PyType_Ready(&VisualType) < 0) return;
	if(PyType_Ready(&PackType)   < 0) return;

	importModules();

	m = Py_InitModule3("_ggimodule", NULL, (char *)NULL);

	Py_INCREF(&ModeType);
	PyModule_AddObject(m, "Mode", (PyObject*)&ModeType);
	Py_INCREF(&PackType);
	PyModule_AddObject(m, "Pack", (PyObject*)&PackType);
	Py_INCREF(&VisualType);
	PyModule_AddObject(m, "Visual", (PyObject*)&VisualType);

	/* export C API */

	capi.ModeType   = &ModeType;
	capi.VisualType = &VisualType; 
	capi.PackType   = &PackType;
	
	capi.attach = pyggi_attachExtension;
	capi.detach = pyggi_detachExtension;

	PyModule_AddObject(m, "_ggimodule_C_api",
			   PyCObject_FromVoidPtr(&capi, NULL));

}
