#include "pygii.h"
#include <ggi/internal/internal.h>
#include <ggi/ggi.h>

typedef struct GGIVisualObject_s Visual;

typedef PyObject *(*ExtGetAttrFunc)(Visual *, char *);
typedef int (*ExtSetAttrFunc)(Visual *, char *, PyObject *);

typedef struct Extension_s {
	char * name;
	ExtGetAttrFunc getattr;
	ExtSetAttrFunc setattr;
	struct Extension_s * next;
} Extension;

struct GGIVisualObject_s {
	PyObject_HEAD
	ggi_visual_t vis;
	ggi_mode     mode;
	const ggi_pixelformat * pf;
	int                     nb_exts;
	struct Extension_s * exts;
};

typedef struct {
	PyObject_HEAD
	ggi_mode mode;
} Mode;

typedef struct {
	PyObject_HEAD
	Visual          * vis;
	ggi_pixel       * pixels;
	size_t          length;
	size_t          alloc;
} Pack;

typedef struct {
	PyTypeObject * VisualType;
	PyTypeObject * ModeType;
	PyTypeObject * PackType;

	int (*attach)(Visual *, char *,  ExtGetAttrFunc ga, ExtSetAttrFunc sa);
	int (*detach)(Visual *, char *);
} pyggi_C_api;


#ifdef PYGGI_MODULE

extern PyTypeObject ModeType;
extern PyTypeObject PackType;
extern PyTypeObject VisualType;

/* Extension mamangement interface */

extern int  pyggi_attachExtension(Visual         *vis,
				 char           *n,
				 ExtGetAttrFunc ga,
				 ExtSetAttrFunc sa);
extern int  pyggi_detachExtension(Visual *vis, char   *name);
extern void pyggi_detachAll(Visual *vis);
extern PyObject * pyggi_getExtensionList(Visual *vis);

/* Helpers */

extern int pyggi_getInt    (PyObject *o, int *i);
extern int pyggi_getCoords (PyObject *o, ggi_coord *c);
extern int pyggi_getRect   (PyObject *o, int i[4]);
extern int pyggi_getPixel  (PyObject *o, ggi_pixel *p, ggi_visual * v);

/* */

extern PyObject * pyggi_newMode(ggi_mode);
extern PyObject * pyggi_newPackFromLength(Visual *, int);
extern PyObject * pyggi_newPackFromPack(Visual *, Pack *);
extern PyObject * pyggi_newPackFromSequence(Visual *, PyObject *);

#   define VisualCheck(op) ((op)->ob_type == &VisualType)
#   define ModeCheck(op)   ((op)->ob_type == &ModeType)
#   define PackCheck(op)   ((op)->ob_type == &PackType)

#else

extern pyggi_C_api * ggi_c_api;

#   define VisualCheck(op) ((op)->ob_type == ggi_c_api->VisualType)
#   define ModeCheck(op)   ((op)->ob_type == ggi_c_api->ModeType)
#   define PackCheck(op)   ((op)->ob_type == ggi_c_api->PackType)

#endif


/* helpers */

#define  float2color(dr,dg,db,c) do { \
    (c)->r=(int)((dr)*65535); \
    (c)->g=(int)((dg)*65535); \
    (c)->b=(int)((db)*65535); } while(0)

#define  color2float(c,dr,dg,db) do { \
    dr=((double)((c)->r))/65535; \
    dg=((double)((c)->g))/65535; \
    db=((double)((c)->b))/65535; } while(0)

#define PACK_HOLDS(p,s)    ( ((Pack*)p)->length >= s  )

#define ASSERT_VISUAL(o,r) do { if(!VisualCheck(o)) { \
	       PyErr_SetString(PyExc_TypeError,"Visual expected"); \
	       return r; } } while(0)
#define ASSERT_VALID_VISUAL(i,r) do { if(i->vis == NULL) { \
	       PyErr_SetString(PyExc_TypeError,"invalid Visual"); \
	       return r; } } while(0)
